/*
  File Name     : ft602_patt_gen.v 
  Department    : IC Design, FTDI SGP
  Author        : Do Ngoc Duong
  History       : 17 Aug. 2016 - Intial version
                  12 Dec. 2016 - Configurable version 

  Description   : This module generates test video patterns  
*/
module ft602_patt_gen (  
  //
  input  rst_n,
  input  sys_clk,
  input [7:0]  mode_select,
  input [15:0] control,
  input [31:0] v_width,
  input [15:0] l_valid,
  input [15:0] l_sg_eg,
  input [15:0] h_blank,
  input [15:0] d_valid,
  input [15:0] d_sg_eg,
  //
  input  fhd_clk,
  input  hd_clk,
  input  sd_clk,  
  //
  output alive,  
  //
  output seq_ena,
  output [31:0] seq_data,
  //  
  output reg gen_ena,  
  output wire gen_pclk,
  output reg gen_vsync,
  output reg gen_hsync,
  output reg gen_de,
  output reg [35:0] gen_data 
);
//**********************PARAMETER**************************
//localparam FRPS = 60;
localparam IGAP      = 4'b0100; 
localparam Y_White   = 8'd180;
localparam Y_Yellow  = 8'd168;
localparam Y_Cyan    = 8'd145;
localparam Y_Green   = 8'd133;
localparam Y_Magenta = 8'd63;
localparam Y_Red     = 8'd51;
localparam Y_Blue    = 8'd28;
localparam Y_Black   = 8'd16;
localparam U_White   = 8'd128;
localparam U_Yellow  = 8'd44;
localparam U_Cyan    = 8'd147;
localparam U_Green   = 8'd63;
localparam U_Magenta = 8'd193;
localparam U_Red     = 8'd109;
localparam U_Blue    = 8'd212;
localparam U_Black   = 8'd128;
localparam V_White   = 8'd128;
localparam V_Yellow  = 8'd136;
localparam V_Cyan    = 8'd44;
localparam V_Green   = 8'd52;
localparam V_Magenta = 8'd204;
localparam V_Red     = 8'd212;
localparam V_Blue    = 8'd120;
localparam V_Black   = 8'd128;
//*********************************************************
// Clock Switching 
reg [1:0] clk_sel;
reg [3:0] clk_cnt; 
reg clk_rstn;
reg selected_clk;
//
always @ (posedge sys_clk or negedge rst_n)
begin 
  if (!rst_n)
    clk_sel <= 2'b00;
  else 
    clk_sel <= control[2:1];
end
// 
always @ (posedge sys_clk or negedge rst_n)
begin 
  if (!rst_n)
    clk_rstn <= 1'b0;
  else if (clk_sel != control[2:1])
    clk_rstn <= 1'b0;
  else if (&clk_cnt)
    clk_rstn <= 1'b1;
end 
//
always @ (posedge sys_clk or negedge rst_n)
begin 
  if (!rst_n)
    clk_cnt <= 'b0;
  else if (clk_sel != control[2:1])
    clk_cnt <= 'b0;
  else if ((!clk_rstn) & (!(&clk_cnt))) 
    clk_cnt <= clk_cnt + 1'b1;
end 
//
always @ *
begin 
  case (clk_sel)
    2'b00:    selected_clk = sd_clk;
    2'b01:    selected_clk = hd_clk;
    2'b10:    selected_clk = fhd_clk;
    default:  selected_clk = sd_clk;
  endcase 
end 
//
wire clk;
assign clk = selected_clk;
//********************************************************* 
//Synchronization
reg [1:0] ftdi_sync_clk_rstn;
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    ftdi_sync_clk_rstn <= 2'b00;
  else 
    ftdi_sync_clk_rstn <= {ftdi_sync_clk_rstn[0],clk_rstn}; 
end
//
wire irst_n;
assign irst_n = ftdi_sync_clk_rstn[1];
//  
reg [7:0]  ftdi_sync_mode;
reg [15:0] ftdi_sync_control;
reg [31:0] ftdi_sync_v_width;
reg [15:0] ftdi_sync_l_valid;
reg [15:0] ftdi_sync_l_sg_eg;
reg [15:0] ftdi_sync_h_blank;
reg [15:0] ftdi_sync_d_valid;
reg [15:0] ftdi_sync_d_sg_eg;
reg [7:0]  reg_mode;
reg [15:0] reg_control;
reg [31:0] reg_v_width;
reg [15:0] reg_l_valid;
reg [15:0] reg_l_sg_eg; 
reg [15:0] reg_h_blank;
reg [15:0] reg_d_valid;
reg [15:0] reg_d_sg_eg; 
//
wire usb_hspeed;
assign usb_hspeed = reg_mode[2];
//
always @ (posedge clk or negedge irst_n)
begin 
  if (!irst_n)
  begin 
    ftdi_sync_mode    <= 'b0;
    ftdi_sync_control <= 'b0;
    ftdi_sync_v_width <= 'b0;
    ftdi_sync_l_valid <= 'b0;
    ftdi_sync_l_sg_eg <= 'b0;
    ftdi_sync_h_blank <= 'b0;
    ftdi_sync_d_valid <= 'b0;
    ftdi_sync_d_sg_eg <= 'b0;
	 reg_mode    <= 'b0;
    reg_control <= 'b0;
    reg_v_width <= 'b0;
    reg_l_valid <= 'b0;
    reg_l_sg_eg <= 'b0;
    reg_h_blank <= 'b0;
    reg_d_valid <= 'b0;
    reg_d_sg_eg <= 'b0; 
  end
  else 
  begin
    ftdi_sync_mode    <= mode_select;
    ftdi_sync_control <= control;
    ftdi_sync_v_width <= v_width;
    ftdi_sync_l_valid <= l_valid;
    ftdi_sync_l_sg_eg <= l_sg_eg;
    ftdi_sync_h_blank <= h_blank;
    ftdi_sync_d_valid <= d_valid;
    ftdi_sync_d_sg_eg <= d_sg_eg;
	 reg_mode    <= ftdi_sync_mode;
    reg_control <= ftdi_sync_control;
    reg_v_width <= ftdi_sync_v_width;
    reg_l_valid <= usb_hspeed ? 16'd240 : ftdi_sync_l_valid;
    reg_l_sg_eg <= ftdi_sync_l_sg_eg;
    reg_h_blank <= usb_hspeed ? 16'd1024 : ftdi_sync_h_blank;
    reg_d_valid <= usb_hspeed ? 16'd320 : ftdi_sync_d_valid;
    reg_d_sg_eg <= ftdi_sync_d_sg_eg;
  end 
end     
//*********************************************************
wire [31:0] real_v_width;
reg [7:0] frame_rate;
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    frame_rate <= 8'h3C;
  else if (reg_control[15:8] == 8'h00)
    frame_rate <= 8'h01;
  else 
    frame_rate <= reg_control[15:8];
end
//	   
`ifdef ALTERA_FPGA 

divider i_div (
  .clock    (clk),
  .denom    (frame_rate),
  .numer    (reg_v_width),
  .quotient (real_v_width),
  .remain   ()
  );
`else

divider i_div (
  .rfd       (),
  .clk       (clk),
  .dividend  (reg_v_width),
  .quotient  (real_v_width),
  .divisor   (frame_rate),
  .fractional()
  );
`endif
// Start Generate Video Frame  
reg [15:0] start_cnt;
wire  pat_ena;
wire  pat_mov;
wire  seq_enable;
reg [7:0] fram_rate; 
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    fram_rate <= 8'd30;
  else 
    case (reg_control[3:1])
      3'b000 : fram_rate <= 8'd30; 
      3'b010 : fram_rate <= 8'd30;
      default: fram_rate <= 8'd30;
    endcase
end 
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    start_cnt <= 16'd0;
  else if (!gen_ena) 
    start_cnt <= 16'd0;
  else if (!start_cnt[8])
    start_cnt <= start_cnt + 1'b1;
end
//
assign pat_ena = reg_control[0] & start_cnt[8];
assign pat_mov = !reg_control[7]; 
assign seq_enable = reg_control[4];
//Select parameters for video formats
wire [1:0] nobar;
assign nobar = reg_control[6:5]; 
//
reg [15:0] hpoint [6:0]; 
// 
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
  begin
    hpoint[0]  <= 'b0; 
    hpoint[1]  <= 'b0; 
    hpoint[2]  <= 'b0; 
    hpoint[3]  <= 'b0; 
    hpoint[4]  <= 'b0; 
    hpoint[5]  <= 'b0; 
    hpoint[6]  <= 'b0; 
  end 
  else
  begin 
    case (nobar)
      2'b00 :
      begin 
        hpoint[0]  <= {reg_d_valid>>3}; 
        hpoint[1]  <= {reg_d_valid>>2}; 
        hpoint[2]  <= {reg_d_valid>>3} + {reg_d_valid>>2}; 
        hpoint[3]  <= {reg_d_valid>>1};
        hpoint[4]  <= {reg_d_valid>>1} + {reg_d_valid>>3};
        hpoint[5]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[6]  <= {reg_d_valid}    - {reg_d_valid>>3};
      end 
      2'b01 :
      begin 
        hpoint[0]  <= {reg_d_valid>>2}; 
        hpoint[1]  <= {reg_d_valid>>2}; 
        hpoint[2]  <= {reg_d_valid>>3} + {reg_d_valid>>2}; 
        hpoint[3]  <= {reg_d_valid>>1};
        hpoint[4]  <= {reg_d_valid>>1} + {reg_d_valid>>3};
        hpoint[5]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[6]  <= {reg_d_valid}    - {reg_d_valid>>3};
      end 
      2'b10 :
      begin 
        hpoint[0]  <= {reg_d_valid>>1}; 
        hpoint[1]  <= {reg_d_valid>>1}; 
        hpoint[2]  <= {reg_d_valid>>1}; 
        hpoint[3]  <= {reg_d_valid>>1};
        hpoint[4]  <= {reg_d_valid>>1} + {reg_d_valid>>3};
        hpoint[5]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[6]  <= {reg_d_valid}    - {reg_d_valid>>3};
      end 
      2'b11 :
      begin 
        hpoint[0]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[1]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[2]  <= {reg_d_valid>>1} + {reg_d_valid>>2}; 
        hpoint[3]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[4]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[5]  <= {reg_d_valid>>1} + {reg_d_valid>>2};
        hpoint[6]  <= {reg_d_valid}    - {reg_d_valid>>3};
      end
    endcase  
  end  
end 
//
reg [9:0] hhpoint0;
reg [9:0] hhpoint1;
reg [9:0] hhpoint2;
reg [9:0] hhpoint3;
reg [9:0] hhpoint4;
reg [9:0] hhpoint5;
reg [9:0] hhpoint6;
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
  begin
    hhpoint0 <= 10'd0;
    hhpoint1 <= 10'd0;
    hhpoint2 <= 10'd0;
    hhpoint3 <= 10'd0;
    hhpoint4 <= 10'd0;
    hhpoint5 <= 10'd0;	 
    hhpoint6 <= 10'd0;	 	 
  end 
  else 
  begin
    hhpoint0 <= hpoint[0][15:6];
    hhpoint1 <= hpoint[1][15:6];
    hhpoint2 <= hpoint[2][15:6];
    hhpoint3 <= hpoint[3][15:6];
    hhpoint4 <= hpoint[4][15:6];
    hhpoint5 <= hpoint[5][15:6];
    hhpoint6 <= hpoint[6][15:6];	 
  end 
end
//    
reg  hsync;
reg  vsync;
reg  valid;
reg  valid_p1;
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    valid_p1 <= 1'b0;
  else 
    valid_p1 <= valid;
end 
//
reg  [7:0]  fram_cnt; 
reg  [31:0] cycf_cnt;
reg  [15:0] line_cnt;
reg  [15:0] cycl_cnt;
reg  [15:0] move_cnt;
wire fram_start;
wire line_start; 
//
assign fram_start = (cycf_cnt == 32'd1); 
assign line_start = (cycl_cnt == 16'd1); 
reg [15:0] reg_h_width;
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    reg_h_width <= 16'd0;
  else 
    reg_h_width <= reg_d_valid + reg_h_blank + reg_d_sg_eg[15:8] + reg_d_sg_eg[7:0];
end 
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
  begin
    fram_cnt <= 8'd0;
    cycf_cnt <= 32'd0;
    line_cnt <= 16'd0;
    cycl_cnt <= 16'd0;
    move_cnt <= 16'd0;
  end 
  else if (!pat_ena)
  begin
    fram_cnt <= 8'd0;
    cycf_cnt <= 32'd0;
    line_cnt <= 16'd0;
    cycl_cnt <= 16'd0;
    move_cnt <= 16'd0;
  end
  else 
  begin
    //Frame count 
    if (fram_start) 
      fram_cnt <= (fram_cnt == fram_rate) ? 6'd1 : fram_cnt + 1'b1;
    //Cycle of frame count  
    cycf_cnt <= (cycf_cnt == real_v_width) ? 32'd1 : cycf_cnt + 1'b1;
    //Line count 
    if (fram_start)
      line_cnt <= 16'd0;
    else if (line_start)
      line_cnt <= line_cnt + 1'b1;  
    //Cycle of line count
    if (vsync | (cycf_cnt < {24'd0,reg_l_sg_eg[15:8]}))
      cycl_cnt <= 16'd0;
    else if (cycl_cnt == reg_h_width)  
      cycl_cnt <= 16'd1;
    else if ((line_cnt <= reg_l_valid) && (cycl_cnt < reg_h_width))
      cycl_cnt <= cycl_cnt + 1'b1;
    //Moving count 
    if (!pat_mov)  
      move_cnt <= 16'd0;
    else if (fram_start && (fram_cnt == fram_rate))
      move_cnt <= (move_cnt == reg_d_valid) ? 16'd0 : move_cnt + 4'd2; 
  end  
end 
//
reg [15:0] vsync_low;
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    vsync_low <= 16'd0;
  else 
    vsync_low <=  reg_d_valid + reg_l_sg_eg[7:0] + reg_d_sg_eg[15:8] + reg_d_sg_eg[7:0];
end
//	   
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    vsync <= 1'b1;
  else if (!pat_ena)
    vsync <= 1'b1;
  else if (fram_start) 
    vsync <= 1'b0;
  else if ((line_cnt == reg_l_valid) & (cycl_cnt == vsync_low)) 
    vsync <= 1'b1;
end 
//
reg [15:0] hsync_low;
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    hsync_low <= 16'd0;
  else 
    hsync_low <= reg_d_valid + reg_d_sg_eg[7:0] + reg_d_sg_eg[15:8];  
end    
//  
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    hsync <= 1'b1;
  else if (!pat_ena)
    hsync <= 1'b1;
  else if ((cycl_cnt == 16'd1) & (line_cnt <= reg_l_valid))
    hsync <= 1'b0;
  else if (cycl_cnt == hsync_low)  
    hsync <= 1'b1;
end   
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    valid <= 1'b0;
  else if (!pat_ena)
    valid <= 1'b0;
  else if ((cycl_cnt == reg_d_sg_eg[15:8]) && (!hsync) & (!vsync))
    valid <= 1'b1;
  else if (cycl_cnt == (reg_d_valid + reg_d_sg_eg[15:8])) 
    valid <= 1'b0;
end 
// 
reg [15:0] vld_cnt;
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    vld_cnt <= 16'd0;
  else if (!valid)
    vld_cnt <= move_cnt; 
  else 
    vld_cnt <= (vld_cnt == reg_d_valid) ? 16'd0 : (vld_cnt + 1'b1);
end
//
reg [15:0] vld_cntp1;
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    vld_cntp1 <= 16'd0;
  else 
    vld_cntp1 <= vld_cnt;
end	 
//
reg [7:0] Y;
reg [7:0] UV;
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
  begin
    Y  <= 8'h00;
    UV <= 8'h00;
  end
  else if (!valid)
  begin
    Y  <= 8'h00;
    UV <= 8'h00;
  end
  else 
    if (vld_cntp1[15:6] < hhpoint0)
    begin 
      Y  <= Y_White;
      UV <= vld_cntp1[0] ? V_White : U_White;
    end 
    else if (vld_cntp1[15:6] < hhpoint1)
    begin 
      Y  <= Y_Yellow;
      UV <= vld_cntp1[0] ? V_Yellow : U_Yellow;
    end 
    else if (vld_cntp1[15:6] < hhpoint2)
    begin 
      Y  <= Y_Cyan;
      UV <= vld_cntp1[0] ? V_Cyan : U_Cyan;
    end 
    else if (vld_cntp1[15:6] < hhpoint3)
    begin 
      Y  <= Y_Green;
      UV <= vld_cntp1[0] ? V_Green : U_Green;
    end 
    else if (vld_cntp1[15:6] < hhpoint4)
    begin 
      Y  <= Y_Magenta;
      UV <= vld_cntp1[0] ? V_Magenta : U_Magenta;
    end 
    else if (vld_cntp1[15:6] < hhpoint5)
    begin 
      Y  <= Y_Red;
      UV <= vld_cntp1[0] ? V_Red : U_Red;
    end 
    else if (vld_cntp1[15:6] < hhpoint6)
    begin 
      Y  <= Y_Blue;
      UV <= vld_cntp1[0] ? V_Blue : U_Blue;
    end
    else  
    begin 
      Y  <= Y_Black;
      UV <= vld_cntp1[0] ? V_Black : U_Black;
    end
end 
// Sequence data output 
reg [32:0] seq_dat; 
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
    seq_dat <= 33'd0;
  else if (valid_p1 & seq_enable)
    seq_dat <= seq_dat + 1'b1;
end 	 
assign seq_ena = seq_enable;
assign seq_data = seq_dat[32:1];
//Output
assign alive     = fram_cnt[3];
assign gen_pclk  = clk; 
//
always @ (posedge clk or negedge irst_n)
begin
  if (!irst_n)
  begin 
    gen_ena   <= 1'b0;
    gen_vsync <= 1'b1;
    gen_hsync <= 1'b1;
    gen_de    <= 1'b0;
    gen_data  <= 36'd0;
  end   
  else
  begin 
    gen_ena   <= reg_control[0]; 
    gen_vsync <= vsync;
    gen_hsync <= hsync;
    gen_de    <= valid_p1;
    gen_data  <= {UV,4'h0,Y,16'h0000};
  end
end  
  
endmodule 
